/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"

#include <algorithm>
#include <cmath>
#include <cstddef>
#include <iterator>
#include <memory>
#include <utility>

#include <Eigen/Geometry>

#include <Ice/Current.h>

#include <SimoxUtility/math/convert/rpy_to_mat3f.h>
#include <VirtualRobot/MathTools.h>
#include <VirtualRobot/Robot.h>

#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/LogSender.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>
#include <ArmarXCore/util/CPPUtility/trace.h>

#include "ArmarXGui/libraries/RemoteGui/Client/Widgets.h"

#include "RobotAPI/components/ArViz/Client/Elements.h"
#include "RobotAPI/components/ArViz/Client/elements/Color.h"
#include "RobotAPI/libraries/armem/core/MemoryID.h"
#include "RobotAPI/libraries/armem/core/Time.h"
#include "RobotAPI/libraries/armem_vision/OccupancyGridHelper.h"
#include "RobotAPI/libraries/armem_vision/client/occupancy_grid/Reader.h"
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

#include "armarx/navigation/algorithms/astar/NavigationCostMap.h"
#include <armarx/navigation/algorithms/astar/util.h>
#include <armarx/navigation/client/NavigationStackConfig.h>
#include <armarx/navigation/client/PathBuilder.h>
#include <armarx/navigation/client/ice/NavigatorInterface.h>
#include <armarx/navigation/client/ice_conversions.h>
#include <armarx/navigation/components/Navigator/RemoteGui.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/factories/NavigationStackFactory.h>
#include <armarx/navigation/server/Navigator.h>
#include <armarx/navigation/server/event_publishing/MemoryPublisher.h>
#include <armarx/navigation/server/execution/PlatformUnitExecutor.h>
#include <armarx/navigation/server/introspection/MemoryIntrospector.h>
#include <armarx/navigation/util/util.h>

namespace armarx::navigation::components
{
    ARMARX_DECOUPLED_REGISTER_COMPONENT(Navigator);
}

namespace armarx::navigation
{

    std::vector<core::Pose>
    convert(const std::vector<Eigen::Matrix4f>& wps)
    {
        using namespace armarx::navigation;
        std::vector<core::Pose> p;
        p.reserve(wps.size());
        std::transform(wps.begin(),
                       wps.end(),
                       std::back_inserter(p),
                       [](const auto& p) { return core::Pose(p); });
        return p;
    }

} // namespace armarx::navigation

namespace armarx::navigation::components
{

    components::Navigator::Navigator() :
        parameterizationReader(memoryNameSystem()),
        parameterizationWriter(memoryNameSystem()),
        eventsWriter(memoryNameSystem()),
        resultsWriter(memoryNameSystem()),
        graphReader(memoryNameSystem()),
        occupancyGridReader(memoryNameSystem()),
        virtualRobotReader(memoryNameSystem()),
        parameterizationService(&parameterizationReader, &parameterizationWriter)
    // publisher(&resultsWriter, &eventsWriter)
    {
        scene.timeServer = &timeServer;
    }


    components::Navigator::~Navigator() = default;

    void
    components::Navigator::onInitComponent()
    {
    }

    void
    components::Navigator::onConnectComponent()
    {
        ARMARX_TRACE;

        static bool initialized{false};

        // redirect to onReconnect to avoid any setup issues
        if (initialized)
        {
            ARMARX_INFO << "Reconnecting ...";
            onReconnectComponent();
            return;
        }

        // connect to all memories
        ARMARX_TRACE;
        resultsWriter.connect();
        eventsWriter.connect();
        // parameterizationReader.connect();
        parameterizationWriter.connect();
        graphReader.connect();
        occupancyGridReader.connect();
        virtualRobotReader.connect();

        // initialize scene
        ARMARX_TRACE;
        updateContext();

        executor = server::PlatformUnitExecutor(platformUnit);

        introspector = server::ArvizIntrospector(getArvizClient(), scene.robot);

        // TODO dynamic scene
        // TODO memory
        // TODO param (10)


        robotStateUpdateTask = new PeriodicTask<Navigator>(
            this, &Navigator::updateRobot, 10, false, "RobotStateUpdateTask");
        robotStateUpdateTask->start();

        navRemoteGui = std::make_unique<NavigatorRemoteGui>(remoteGui, *this);
        navRemoteGui->enable();

        initialized = true;

        ARMARX_INFO << "Initialized. Will now respond to navigation requests.";
    }

    void
    components::Navigator::onReconnectComponent()
    {
        ARMARX_TRACE;

        robotStateUpdateTask->start();

        // TODO not in all cases meaningful
        //resume();
        resultsWriter.connect();
        eventsWriter.connect();
        // parameterizationReader.connect();
        parameterizationWriter.connect();
        graphReader.connect();
        virtualRobotReader.connect();
        occupancyGridReader.connect();

        navRemoteGui->enable();
    }

    void
    components::Navigator::onDisconnectComponent()
    {
        ARMARX_TRACE;

        robotStateUpdateTask->stop();

        stopAll();

        navRemoteGui->disable();
    }

    void
    components::Navigator::onExitComponent()
    {
    }

    void
    components::Navigator::updateContext()
    {
        ARMARX_TRACE;

        scene.robot = getRobot();

        {
            std::lock_guard g{scene.staticSceneMtx};
            scene.staticScene = staticScene();
        }

        // TODO dynamic scene
        {
            std::lock_guard g{scene.graphMtx};
            scene.graph = core::SceneGraph{.subgraphs = graphReader.graphs()};
        }
    }

    std::string
    components::Navigator::getDefaultName() const
    {
        return "Navigator";
    }

    void
    components::Navigator::createConfig(const aron::data::dto::DictPtr& stackConfig,
                                        const std::string& callerId,
                                        const Ice::Current&)
    {
        // TODO: Not thread-safe.
        ARMARX_TRACE;
        ARMARX_INFO << "Creating config for caller '" << callerId << "'";

        parameterizationService.store(stackConfig, callerId, timeServer.now());

        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);

        memoryIntrospectors.emplace_back(
            std::make_unique<server::MemoryIntrospector>(resultsWriter, callerId));

        // keep track of all memory publishers
        memoryPublishers.emplace(
            callerId,
            std::make_unique<server::MemoryPublisher>(&resultsWriter, &eventsWriter, callerId));

        navigators.emplace(
            std::piecewise_construct,
            std::forward_as_tuple(callerId),
            std::forward_as_tuple(
                server::Navigator::Config{.stack = std::move(stack),
                                          .scene = &scene,
                                          .general = server::Navigator::Config::General{}},
                server::Navigator::InjectedServices{.executor = &executor.value(),
                                                    .publisher =
                                                        memoryPublishers.at(callerId).get(),
                                                    .introspector = &(introspector.value())}));
    }

    void
    components::Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
                                  const std::string& navigationMode,
                                  const std::string& callerId,
                                  const Ice::Current&)
    {
        ARMARX_TRACE;

        ARMARX_INFO << "moveTo() requested by caller '" << callerId << "'";
        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        try
        {
            updateContext();
            navigators.at(callerId).moveTo(convert(waypoints),
                                           core::NavigationFrameNames.from_name(navigationMode));
        }
        catch (...)
        {
            // TODO: Error handling.
            ARMARX_WARNING << "Failed to execute moveTo()."
                           << "Movement will be paused.";
            stopAll(); // TODO pause movement must not throw
            throw;
        }
    }


    void
    Navigator::moveTo2(const client::detail::Waypoints& waypoints,
                       const std::string& navigationMode,
                       const std::string& callerId,
                       const Ice::Current& c)
    {
        ARMARX_TRACE;

        ARMARX_IMPORTANT << "moveTo2 requested by caller '" << callerId << "'";

        const std::vector<client::WaypointTarget> wps = client::defrost(waypoints);

        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        updateContext();
        navigators.at(callerId).moveTo(wps, core::NavigationFrameNames.from_name(navigationMode));

        // navigators.at(callerId).moveTo()
    }

    void
    components::Navigator::moveTowards(const Eigen::Vector3f& direction,
                                       const std::string& navigationMode,
                                       const std::string& callerId,
                                       const Ice::Current&)
    {
        // TODO: Error handling.
        ARMARX_TRACE;
        ARMARX_INFO << "MoveTowards requested by caller '" << callerId << "'";

        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        updateContext();
        navigators.at(callerId).moveTowards(direction,
                                            core::NavigationFrameNames.from_name(navigationMode));
    }

    void
    components::Navigator::pause(const std::string& configId, const Ice::Current&)
    {
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).pause();
    }

    void
    components::Navigator::resume(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).resume();
    }

    void
    components::Navigator::stop(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).stop();
    }

    void
    components::Navigator::stopAll(const Ice::Current&)
    {
        ARMARX_IMPORTANT << "stopAll()";

        ARMARX_TRACE;
        for (auto& [_, navigator] : navigators)
        {
            navigator.stop();
        }

        const core::GoalReachedEvent evt{{.timestamp = scene.timeServer->now()},
                                         {.pose = core::Pose(scene.robot->getGlobalPose())}};

        for (auto& [callerId, memoryPublisher] : memoryPublishers)
        {
            // TODO use different event, e.g. generic stopped event
            // TODO check if navigator was stopped
            memoryPublisher->goalReached(evt);
        }
    }

    bool
    components::Navigator::isPaused(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        return navigators.at(configId).isPaused();
    }

    bool
    components::Navigator::isStopped(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        return navigators.at(configId).isStopped();
    }

    armarx::PropertyDefinitionsPtr
    components::Navigator::createPropertyDefinitions()
    {
        ARMARX_TRACE;

        PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        def->component(platformUnit);

        def->component(remoteGui, "RemoteGuiProvider");
        // Add a required property.
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        def->optional(params.occupiedGridThreshold,
                      "p.occupancy_grid.occopied_threshold",
                      "Threshold for each cell to be considered occupied. Increase this value to "
                      "reduce noise.");

        resultsWriter.registerPropertyDefinitions(def);
        eventsWriter.registerPropertyDefinitions(def);
        // parameterizationReader.registerPropertyDefinitions(def);
        parameterizationWriter.registerPropertyDefinitions(def);
        graphReader.registerPropertyDefinitions(def);
        virtualRobotReader.registerPropertyDefinitions(def);
        occupancyGridReader.registerPropertyDefinitions(def);

        return def;
    }

    core::StaticScene
    components::Navigator::staticScene()
    {
        ARMARX_TRACE;

        const objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();
        core::StaticScene scene{.objects = util::asSceneObjects(objectPoses)};

        ARMARX_INFO << "The object scene consists of " << scene.objects->getSize() << " objects";

        const armem::vision::occupancy_grid::client::Reader::Result result =
            occupancyGridReader.query(armem::vision::occupancy_grid::client::Reader::Query{
                .providerName = "CartographerMappingAndLocalization",
                .timestamp = armem::Time::now()});

        if (result and result.occupancyGrid.has_value() and false) // Feature disabled on master ...
        {
            ARMARX_INFO << "Occupancy grid available!";

            const auto occupancyGridSceneElements = util::asSceneObjects(
                result.occupancyGrid.value(),
                OccupancyGridHelper::Params{.freespaceThreshold = 0.45F,
                                            .occupiedThreshold = params.occupiedGridThreshold});
            ARMARX_INFO << occupancyGridSceneElements->getSize()
                        << " scene elements from occupancy grid";

            scene.objects->addSceneObjects(occupancyGridSceneElements);

            // draw
            auto layer = arviz.layer("occupancy_grid");

            for (const auto& sceneObject : occupancyGridSceneElements->getSceneObjects())
            {
                const Eigen::Isometry3f world_T_obj(sceneObject->getGlobalPose());
                ARMARX_INFO << world_T_obj.translation();
                ARMARX_INFO << layer.size();
                layer.add(viz::Box("box_" + std::to_string(layer.size()))
                              .pose(world_T_obj)
                              .size(result.occupancyGrid->resolution)
                              .color(viz::Color::orange()));
            }

            ARMARX_INFO << "Creating costmap";

            algorithm::astar::NavigationCostMap costmap(
                getRobot(),
                scene.objects,
                algorithm::astar::NavigationCostMap::Parameters{.binaryGrid = false,
                                                                .cellSize = 100});

            costmap.createCostmap();

            ARMARX_INFO << "Done";

            ARMARX_TRACE;
            const auto grid = costmap.getGrid();

            ARMARX_TRACE;
            ARMARX_INFO << "Dumping.";
            dumpToFile(grid);

            arviz.commit({layer});
        }
        else
        {
            ARMARX_INFO << "Occupancy grid not available";
        }

        return scene;
    }

    VirtualRobot::RobotPtr
    components::Navigator::getRobot()
    {
        ARMARX_TRACE;

        auto robot = RemoteRobot::createLocalCloneFromFile(
            getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
        // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
        ARMARX_CHECK_NOT_NULL(robot);

        // ARMARX_INFO << "Resetting robot";
        // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

        RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

        return robot;
    }

    void
    components::Navigator::updateRobot()
    {
        ARMARX_TRACE;
        synchronizeLocalClone(scene.robot);

        auto robotDescription = virtualRobotReader.queryDescription("Armar6", armem::Time::now());

        ARMARX_CHECK(robotDescription.has_value());

        // synchronizing platform state
        const std::optional<armem::robot::PlatformState> platformState =
            virtualRobotReader.queryPlatformState(robotDescription.value(), armem::Time::now());

        ARMARX_CHECK(platformState.has_value());

        // TODO / FIXME make this thread safe!
        {
            std::lock_guard g{scene.platformVelocityMtx};

            if (not scene.platformVelocity.has_value())
            {
                scene.platformVelocity = core::Twist();
            }

            scene.platformVelocity->linear = platformState->twist.linear;
            scene.platformVelocity->angular = platformState->twist.angular;
        }
    }

    server::Navigator*
    components::Navigator::activeNavigator()
    {
        ARMARX_TRACE;
        // We define the active navigator to be the one whose movement is enabled.
        const auto isActive = [](auto& nameNavPair) -> bool
        {
            server::Navigator& navigator = nameNavPair.second;
            return not navigator.isPaused();
        };

        auto it = std::find_if(navigators.begin(), navigators.end(), isActive);

        // no navigator active?
        if (it == navigators.end())
        {
            return nullptr;
        }

        return &it->second;
    }

} // namespace armarx::navigation::components
